{% extends "base.html" %}
{% block heading %}
Tree {{ name | e }}
{% endblock %}

{% block content %}
<h1>Tree {{ name | e }}</h1>

<ul>
  {% for entry in entries %}
  <li>
    <a href="/{{ entry["type"] | e}}/{{ entry["hash"] | e}}">{{ entry["path"] | e}}</a>
    {% if entry["type"] == "blob" %}
    <a href="/blob/{{ entry["hash"] | e}}/{{ entry["path"] | e}}">[&darr;]</a>
    {% endif %}
  </li>
  {% endfor %}
</ul>
{% endblock %}
