<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN">
<html>
  <head>
    <title>
      {% block heading %}Invocation Browser{% endblock %}
    </title>
    <link rel="stylesheet" type="text/css" href="/static/style.css">
  </head>
  <body>
    {% block content %}
    <i>(no content)</i>
    {% endblock %}
  </body>
</html>
