// Copyright 2022 Huawei Cloud Computing Technology Co., Ltd.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#pragma once

#include <iostream>

class Printer {
  public:
    template <class SequenceT>
    void print(SequenceT& seq, unsigned int number_of_terms) {
        if (number_of_terms == 0) {
            std::cout << std::endl;
            return;
        }
        for (unsigned int i = 0; i < number_of_terms - 1; ++i) {
            std::cout << seq.next() << seq.separator();
        }
        std::cout << seq.next() << std::endl;
    }
};
