#!/usr/bin/env python3
# Copyright 2025 Huawei Cloud Computing Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

outname = sys.argv[1]
depth = int(sys.argv[2])

outpath = os.path.join(outname, *(["x"] * depth))

os.makedirs(outpath)

for i in range(10):
    with open(os.path.join(outpath, "data%d.txt" % (i,)), "w") as f:
        f.write("%d" % (i,))
