// Copyright 2024 Huawei Cloud Computing Technology Co., Ltd.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "src/utils/cpp/prefix.hpp"

#include <string>

#include "catch2/catch_test_macros.hpp"

TEST_CASE("lines", "[prefix]") {
    CHECK(PrefixLines("").empty());
    CHECK(PrefixLines("foo\nbar\n") == "  foo\n  bar\n");
    CHECK(PrefixLines("foo\nbar\n", ">") == ">foo\n>bar\n");
    CHECK(PrefixLines("foo") == "  foo\n");
    CHECK(PrefixLines("foo\nbar") == "  foo\n  bar\n");
}
