// Copyright 2022 Huawei Cloud Computing Technology Co., Ltd.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef INCLUDED_SRC_BUILDTOOL_COMMON_PROTOCOL_TRAITS_HPP
#define INCLUDED_SRC_BUILDTOOL_COMMON_PROTOCOL_TRAITS_HPP

#include "src/buildtool/crypto/hash_function.hpp"

class ProtocolTraits final {
  public:
    static constexpr auto IsNative(HashFunction::Type hash_type) noexcept
        -> bool {
        return hash_type == HashFunction::Type::GitSHA1;
    }

    static constexpr auto IsTreeAllowed(HashFunction::Type hash_type) noexcept
        -> bool {
        return IsNative(hash_type);
    }
};
#endif  // INCLUDED_SRC_BUILDTOOL_COMMON_PROTOCOL_TRAITS_HPP
