#!/usr/bin/env python3
# Copyright 2025 Huawei Cloud Computing Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

with open("out/failures.json") as f:
    failures = json.load(f)

with open("format.diff", "w") as f:
    for k, v in failures.items():
        src = v["config"]["src"]
        if src.startswith("work/"):
            src = src[len("work/"):]
        diff = v["log"].splitlines()
        f.write("--- a/%s\n" % (src, ))
        f.write("+++ b/%s\n" % (src, ))
        for l in diff[2:]:
            f.write(l + "\n")
